@@ -1,4 +1,4 @@
-# $Id: Client-Keepalive.pm 73 2006-09-26 17:57:54Z rcaputo $
+# $Id: Client-Keepalive.pm 83 2008-03-24 16:37:59Z rcaputo $
 
 package POE::Component::Client::Keepalive;
 
@@ -6,7 +6,7 @@ use warnings;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = "0.0901";
+$VERSION = "0.1001";
 
 use Carp qw(croak);
 use Errno qw(ETIMEDOUT);
@@ -59,6 +59,7 @@ sub SF_TIMEOUT   () {  9 }   #   $default_request_timeout,
 sub SF_RESOLVER  () { 10 }   #   $poco_client_dns_object,
 sub SF_SHUTDOWN  () { 11 }   #   $shutdown_flag,
 sub SF_REQ_INDEX () { 12 }   #   \%request_id_to_wheel_id,
+sub SF_BIND_ADDR () { 13 }   #   $bind_address,
                              # );
 
                             # $socket_xref{$socket} = [
@@ -116,6 +117,7 @@ sub new {
   my $keep_alive   = delete($args{keep_alive})   || 15;
   my $timeout      = delete($args{timeout})      || 120;
   my $resolver     = delete($args{resolver});
+  my $bind_address = delete($args{bind_address}) || "0.0.0.0";
 
   my @unknown = sort keys %args;
   if (@unknown) {
@@ -135,6 +137,8 @@ sub new {
     $timeout,           # SF_TIMEOUT
     undef,              # SF_RESOLVER
     undef,              # SF_SHUTDOWN
+    undef,              # SF_REQ_INDEX
+    $bind_address,      # SF_BIND_ADDR
   ], $class;
 
   unless (defined $resolver) {
@@ -148,7 +152,7 @@ sub new {
     object_states => [
       $self => {
         _start               => "_ka_initialize",
-        _stop                => "_ka_ignore_this_event",
+        _stop                => "_ka_stopped",
         ka_conn_failure      => "_ka_conn_failure",
         ka_conn_success      => "_ka_conn_success",
         ka_reclaim_socket    => "_ka_reclaim_socket",
@@ -179,6 +183,17 @@ sub _ka_initialize {
   $kernel->alias_set("$object");
 }
 
+# When programs crash, the session may stop in a non-shutdown state.
+# _ka_stopped and DESTROY catch this either way the death occurs.
+
+sub _ka_stopped {
+	$_[OBJECT][SF_SHUTDOWN] = 1;
+}
+
+sub DESTROY {
+  my $self = shift;
+  $self->shutdown();
+}
 
 # Request to wake up.  This should only happen during the edge
 # condition where the component's request queue goes from empty to
@@ -258,6 +273,7 @@ sub _ka_wake_up {
 
     my $addr = ($request->[RQ_IP] or $request->[RQ_ADDRESS]);
     my $wheel = POE::Wheel::SocketFactory->new(
+      BindAddress   => $self->[SF_BIND_ADDR],
       RemoteAddress => $addr,
       RemotePort    => $request->[RQ_PORT],
       SuccessEvent  => "ka_conn_success",
@@ -373,7 +389,7 @@ sub allocate {
     1,          # RQ_ACTIVE
     _allocate_req_id(), # RQ_ID
   ];
-  
+
   $self->[SF_REQ_INDEX]{$request->[RQ_ID]} = $request;
 
   $poe_kernel->refcount_increment(
@@ -712,6 +728,7 @@ sub _ka_relinquish_socket {
 
 sub shutdown {
   my $self = shift;
+  return if $self->[SF_SHUTDOWN];
   $poe_kernel->call("$self", "ka_shutdown");
 }
 
@@ -767,7 +784,7 @@ sub _shutdown_request {
   if (defined $request->[RQ_WHEEL_ID]) {
     DEBUG and warn "SHT: Shutting down resolver wheel $request->[RQ_TIMER_ID]";
     delete $self->[SF_WHEELS]{$request->[RQ_WHEEL_ID]};
-    
+
     # remove the wheel-to-request index
     delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
   }
@@ -921,7 +938,7 @@ sub _ka_add_to_queue {
   return if (
     ($self->[SF_USED_EACH]{$conn_key} || 0) >= $self->[SF_MAX_HOST]
   );
-  
+
   # Wake the session up, and return nothing, signifying sound and fury
   # yet to come.
   DEBUG and warn "posting wakeup for $conn_key";
@@ -1098,7 +1115,7 @@ others.  None will know about the limits set in the others, so it's
 possible to overrun your file descriptors for a process if you're not
 careful.
 
-new() takes up to four parameters.  All of them are optional.
+new() takes up to five parameters.  All of them are optional.
 
 To limit the number of simultaneous connections to a particular host
 (defined by a combination of scheme, address and port):
@@ -1123,6 +1140,11 @@ component holds a request before generating an error.
 
   timeout      => $seconds_to_process_a_request, # defaults to 120
 
+Specify a bind_address to bind all client sockets to a particular
+local address.  The value of bind_address will be passed directly to
+POE::Wheel::SocketFactory.  See that module's documentation for
+implementation details.
+
 =item allocate
 
 Allocate a new connection.  Allocate() will return immediately.  The
@@ -1255,14 +1277,19 @@ sending mail to bug-poe-component-client-keepalive@rt.cpan.org.
 
 =head1 LICENSE
 
-This distribution is copyright 2004 by Rocco Caputo.  All rights are
-reserved.  This distribution is free software; you may redistribute it
-and/or modify it under the same terms as Perl itself.
+This distribution is copyright 2004-2006 by Rocco Caputo.  All rights
+are reserved.  This distribution is free software; you may
+redistribute it and/or modify it under the same terms as Perl itself.
 
 =head1 AUTHOR
 
 Rocco Caputo <rcaputo@cpan.org>
 
-Special thanks to Rob Bloodgood.
+=head1 CONTRIBUTORS
+
+Rob Bloodgood helped out a lot.  Thank you.
+
+Joel Bernstein solved some nasty race conditions.  Portugal Telecom
+L<http://www.sapo.pt/> was kind enough to support his contributions.
 
 =cut
